package edu.fiu.jtlex.timeml.io;


import java.io.InputStream;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Scanner;
import java.util.Set;
import java.util.StringTokenizer;


import edu.fiu.jtlex.timeml.data.TimeMLEvent;
import edu.fiu.jtlex.timeml.data.ITimeMLEvent;
import edu.fiu.jtlex.timeml.data.ITimeMLInstance;
import edu.fiu.jtlex.timeml.data.ITimeMLLink.ITimeMLLinkType;
import edu.fiu.jtlex.timeml.data.ITimeMLSignal;
import edu.fiu.jtlex.timeml.data.ITimeMLLink;
import edu.fiu.jtlex.timeml.data.ITimeMLText;
import edu.fiu.jtlex.timeml.data.ITimeMLTimex;
import edu.fiu.jtlex.timeml.data.TimeMLInstance;
import edu.fiu.jtlex.timeml.data.TimeMLLink;
import edu.fiu.jtlex.timeml.data.TimeMLSignal;
import edu.fiu.jtlex.timeml.data.TimeMLText;
import edu.fiu.jtlex.timeml.data.TimeMLTimex;
import edu.fiu.jtlex.timeml.data.ITimeMLEvent.EventClass;
import edu.fiu.jtlex.timeml.data.ITimeMLInstance.Aspect;
import edu.fiu.jtlex.timeml.data.ITimeMLInstance.Polarity;
import edu.fiu.jtlex.timeml.data.ITimeMLInstance.Pos;
import edu.fiu.jtlex.timeml.data.ITimeMLInstance.Tense;
import edu.fiu.jtlex.timeml.data.ITimeMLLink.ALink;
import edu.fiu.jtlex.timeml.data.ITimeMLLink.Origintype;
import edu.fiu.jtlex.timeml.data.ITimeMLLink.SLink;
import edu.fiu.jtlex.timeml.data.ITimeMLLink.TLink;
import edu.fiu.jtlex.timeml.data.ITimeMLTimex.FunctionInDocument;
import edu.fiu.jtlex.timeml.data.ITimeMLTimex.TimexMod;
import edu.fiu.jtlex.timeml.data.ITimeMLTimex.TimexType;
import edu.fiu.jtlex.timeml.graph.ITimeMLNode;

/**
 * Default implementation of {@code ITimeMLParser}.
 *
 * @since jTLEX 1.0
 * @author arada002
 */
public class TimeMLParser implements ITimeMLParser {

    private String rawText = "";
    private String text = "";
    private Map<Integer, ITimeMLEvent> events = new HashMap<>();
    private Map<Integer, ITimeMLTimex> timexes = new HashMap<>();
    private Map<Integer, ITimeMLSignal> signals = new HashMap<>();
    private Map<Integer, ITimeMLInstance> instances = new HashMap<>();
    private Set<ITimeMLLink> links = new HashSet<ITimeMLLink>();

    /**
     * A method for parsing a TimeML File and returning the found set of events/timex's/signals/instances/links
     * and returns them as a text object.
     *
     * @param tmlFile An InputStream containing the TimeML File.
     * @return An IText object containing the text/raw text/events/timex's/signals/instances/links.
     * @throws NullPointerException if tmlFile argument is <code>null</code>
     */
    public ITimeMLText parse(InputStream tmlFile) {

        Scanner scnr = new Scanner(tmlFile);

        while (scnr.hasNext()) {
            rawText += scnr.next();
            rawText += " ";
        }
        scnr.close();


        for (int i = 0; i < rawText.length(); i++) {

            if (rawText.charAt(i) == '<') {
                if (rawText.charAt(i + 1) == '/') {
                    String endTag = "";
                    while (rawText.charAt(i) != '>' && i <= rawText.length() - 1) {
                        endTag += String.valueOf(rawText.charAt(i));
                        i++;
                    }
                    endTag += String.valueOf(rawText.charAt(i));
                    if (endTag.contains("TimeML")) {
                        break;
                    }

                } else {
                    String tag = "";
                    String tagText = "";
                    i++;
                    while (rawText.charAt(i) != ' ' && rawText.charAt(i) != '>' && i <= rawText.length() - 1) {
                        tag += String.valueOf(rawText.charAt(i));
                        i++;
                    }

                    if (rawText.charAt(i) == '>') {
                        continue;
                    } else {
                        i++;
                        while (rawText.charAt(i) != '>' && i <= rawText.length() - 1) {
                            tagText += String.valueOf(rawText.charAt(i));
                            i++;
                        }
                        tagText += " ";
                        i++;
                        if (tag.equals("?xml") || tag.equals("TimeML")) {
                            i--;
                        } else {
                            while (rawText.charAt(i) != '<' && i <= rawText.length() - 1) {
                                text += String.valueOf(rawText.charAt(i));
                                tagText += String.valueOf(rawText.charAt(i));
                                i++;
                            }
                            i--;

                            if (tag.equals("EVENT")) {
                                if (rawText.substring(i + 2, i + 12).equals("NUMEX TYPE")) {
                                    i = i + 12;
                                    while (rawText.charAt(i) != '>') {
                                        i++;
                                    }

                                    while (rawText.charAt(i) != '<') {
                                        text += String.valueOf(rawText.charAt(i));
                                        tagText += String.valueOf(rawText.charAt(i));
                                        i++;
                                    }
                                    i--;
                                } else if (rawText.substring(i + 2, i + 13).equals("ENAMEX TYPE")) {
                                    i = i + 12;
                                    while (rawText.charAt(i) != '>') {
                                        i++;
                                    }

                                    while (rawText.charAt(i) != '<') {
                                        text += String.valueOf(rawText.charAt(i));
                                        tagText += String.valueOf(rawText.charAt(i));
                                        i++;
                                    }
                                    i--;
                                } else if (rawText.substring(i + 1, i + 11).equals("<CARDINAL>")) {
                                    i = i + 11;
                                    while (rawText.charAt(i) != '<') {
                                        text += String.valueOf(rawText.charAt(i));
                                        tagText += String.valueOf(rawText.charAt(i));
                                        i++;
                                    }
                                    i--;
                                }

                                if (rawText.substring(i + 1, i + 9).equals("<VG-VBG>") || rawText.substring(i + 1, i + 9).equals("<VG-VBN>")) {
                                    i = i + 9;
                                    while (rawText.charAt(i) != '<') {
                                        text += String.valueOf(rawText.charAt(i));
                                        tagText += String.valueOf(rawText.charAt(i));
                                        i++;
                                    }
                                    i--;
                                }
                                if (rawText.substring(i + 1, i + 5).equals("<VG>") || rawText.substring(i + 1, i + 5).equals("<NG>") || rawText.substring(i + 1, i + 5).equals("<JG>") || rawText.substring(i + 1, i + 5).equals("<PG>")) {
                                    i = i + 5;

                                    while (rawText.charAt(i) != '<') {
                                        text += String.valueOf(rawText.charAt(i));
                                        tagText += String.valueOf(rawText.charAt(i));
                                        i++;
                                    }
                                    i--;
                                }

                                if (rawText.substring(i + 1, i + 7).equals("<HEAD>")) {
                                    i = i + 7;
                                    while (rawText.charAt(i) != '<') {
                                        text += String.valueOf(rawText.charAt(i));
                                        tagText += String.valueOf(rawText.charAt(i));
                                        i++;
                                    }
                                    i--;
                                }
                                ITimeMLEvent event = parseEvent(tagText);
                                events.put(event.getId(), event);
                            } else if (tag.equals("TIMEX3")) {
                                if (rawText.substring(i + 1, i + 11).equals("<CARDINAL>")) {
                                    i = i + 11;
                                    while (rawText.charAt(i) != '<') {
                                        text += String.valueOf(rawText.charAt(i));
                                        tagText += String.valueOf(rawText.charAt(i));
                                        i++;
                                    }
                                    i--;
                                }
                                ITimeMLTimex timex = parseTimex(tagText);
                                timexes.put(timex.getId(),timex);
                            } else if (tag.equals("SIGNAL")) {
                                if (rawText.substring(i + 1, i + 11).equals("<CARDINAL>")) {
                                    i = i + 11;
                                    while (rawText.charAt(i) != '<') {
                                        text += String.valueOf(rawText.charAt(i));
                                        tagText += String.valueOf(rawText.charAt(i));
                                        i++;
                                    }
                                    i--;
                                }
                                if (rawText.substring(i + 1, i + 5).equals("<PG>") || rawText.substring(i + 1, i + 5).equals("<RG>") || rawText.substring(i + 1, i + 5).equals("<VG>")) {
                                    i = i + 5;
                                    while (rawText.charAt(i) != '<') {
                                        text += String.valueOf(rawText.charAt(i));
                                        tagText += String.valueOf(rawText.charAt(i));
                                        i++;
                                    }
                                    i--;
                                }

                                ITimeMLSignal signal = parseSignal(tagText);
                                signals.put(signal.getID(), signal);
                            } else if (tag.equals("MAKEINSTANCE")) {
                            	ITimeMLInstance instance = parseInstance(tagText);
                                instances.put(instance.getId(), instance);
                            } else if (tag.contains("LINK")) {
                                links.add(parseLink(tag, tagText));
                            }
                        }
                    }
                }
            } else {
                text += String.valueOf(rawText.charAt(i));
            }
        }

        return new TimeMLText(rawText, text, timexes,
                links, instances, events);
    }

    /**
     * A method for parsing events from a string.
     *
     * @param eventString A string containing an event.
     * @return an IEvent object containing the event's eid, eventClass, and stem.
     * @throws NullPointerException if null
     * @since jTLEX 1.0
     */
    public ITimeMLEvent parseEvent(String eventString) {
        int eid = 0;
        EventClass eClass = null;
        String stem = "";
        StringTokenizer tok = new StringTokenizer(eventString);
        while (tok.hasMoreTokens()) {
            String token = tok.nextToken();
            if (token.contains("eid=\"")) {
                eid = Integer.parseInt(token.substring(token.indexOf('"') + 2, token.length() - 1));
            } else if (token.contains("class=\"")) {
                eClass = EventClass.valueOf(token.substring(token.indexOf('"') + 1, token.length() - 1));
            } else if (token.contains("stem=\"")) {
                if (token.indexOf('"') != token.lastIndexOf('"')) {
                    stem += token.substring(token.indexOf('"') + 1, token.lastIndexOf('"')) + " ";
                } else {
                    stem += token.substring(token.indexOf('"') + 1) + " ";
                }
            } else {
                stem += token + " ";
            }
        }

        return new TimeMLEvent(eid, eClass, stem);
    }

    /**
     * A method for parsing a timex from a string.
     *
     * @param timexString A string containing a timex.
     * @return an Timex object containing the timex's id, type, value, mod, temporal function,
     * anchor Id, document function, beginning point, end point, quant, frequency, and phrase.
     * @throws NullPointerException if null
     * @since jTLEX 1.0
     */
    public ITimeMLTimex parseTimex(String timexString) {
        int tid = 0;
        TimexType type = null;
        String value = "";
        TimexMod mod = null;
        boolean temporalFunction = true;
        int anchorId = 0;
        FunctionInDocument documentFunction = null;
        int beginPoint = 0;
        int endPoint = 0;
        String quant = null;
        String freq = null;
        String phrase = "";

        StringTokenizer tok = new StringTokenizer(timexString);

        while (tok.hasMoreTokens()) {
            String token = tok.nextToken();
            if (token.contains("tid=\"")) {
                tid = Integer.parseInt(token.substring(token.indexOf('"') + 2, token.length() - 1));
            } else if (token.contains("type=\"")) {
                type = TimexType.valueOf(token.substring(token.indexOf('"') + 1, token.length() - 1));
            } else if (token.contains("value=\"")) {
                if (token.indexOf('"') != token.lastIndexOf('"')) {
                    value += token.substring(token.indexOf('"') + 1, token.lastIndexOf('"')) + " ";
                } else {
                    value += token.substring(token.indexOf('"') + 1) + " ";
                }
            } else if (token.contains("mod=\"")) {
                mod = TimexMod.valueOf(token.substring(token.indexOf('"') + 1, token.length() - 1));
            } else if (token.contains("temporalFunction=\"")) {
                temporalFunction = Boolean.parseBoolean(token.substring(token.indexOf('"') + 1, token.lastIndexOf('"')));
            } else if (token.contains("anchorTimeID=\"")) {
                anchorId = Integer.parseInt(token.substring(token.indexOf('"') + 2, token.length() - 1));
            } else if (token.contains("functionInDocument=\"")) {
                documentFunction = FunctionInDocument.valueOf(token.substring(token.indexOf('"') + 1, token.length() - 1));
            } else if (token.contains("beginPoint=\"")) {
                beginPoint = Integer.parseInt(token.substring(token.indexOf('"') + 2, token.length() - 1));
            } else if (token.contains("endPoint=\"")) {
                endPoint = Integer.parseInt(token.substring(token.indexOf('"') + 2, token.length() - 1));
            } else if (token.contains("freq=\"")) {
                freq = token.substring(token.indexOf('"') + 1, token.lastIndexOf('"')) + " ";
            } else if (token.contains("quant=\"")) {
                quant = token.substring(token.indexOf('"') + 1, token.lastIndexOf('"')) + " ";
            } else {
                phrase += token + " ";
            }
        }

        return new TimeMLTimex(tid, type, value, mod, temporalFunction, anchorId, documentFunction,
                beginPoint, endPoint, quant, freq, phrase);
    }

    /**
     * A method for parsing a link from a string.
     *
     * @param tag        a string containing the type of link
     * @param linkString a string containing the link
     * @return an ILink object containing it's link tag, link id, syntax, first node, signal, second node, and relType.
     * @throws NullPointerException if null
     * @since jTLEX 1.0
     */
    public ITimeMLLink parseLink(String tag, String linkString) {
        int lid = 0;
        String syntax = "";
        ITimeMLNode firstNode = null;
        ITimeMLSignal signal = null;
        ITimeMLNode secondNode = null;
        Origintype origin = null;
        Class<?> linkTag = null;
        ITimeMLLinkType relType = null;

        if (tag.contains("ALINK")) {
            linkTag = ITimeMLLink.ALink.class;
        } else if (tag.contains("SLINK")) {
            linkTag = ITimeMLLink.SLink.class;
        } else if (tag.contains("TLINK")) {
            linkTag = ITimeMLLink.TLink.class;
        }
        StringTokenizer tok = new StringTokenizer(linkString);
        while (tok.hasMoreTokens()) {
            String token = tok.nextToken();

            if (token.contains("lid=\"")) {
                lid = Integer.parseInt(token.substring(token.indexOf('"') + 2, token.length() - 1));
            } else if (token.contains("syntax=\"")) {
                syntax = token.substring(token.indexOf('"') + 1, token.lastIndexOf('"'));
            } else if (token.contains("eventInstanceID=\"")) {
                int eiid = Integer.parseInt(token.substring(token.indexOf('"') + 3, token.length() - 1));
                firstNode = instances.get(eiid);
            } else if (token.contains("timeID=\"")) {
                int tid = Integer.parseInt(token.substring(token.indexOf('"') + 2, token.length() - 1));
                firstNode = timexes.get(tid);
            } else if (token.contains("relatedToEventInstance=\"") || token.contains("subordinatedEventInstance=\"")) {
                int eiid = Integer.parseInt(token.substring(token.indexOf('"') + 3, token.lastIndexOf('"')));
                secondNode = instances.get(eiid);
            } else if (token.contains("relatedToTime=\"")) {
                int tid = Integer.parseInt(token.substring(token.indexOf('"') + 2, token.lastIndexOf('"')));
                secondNode = timexes.get(tid);
            } else if (token.contains("signalID=\"")) {
                int sid = Integer.parseInt(token.substring(token.indexOf('"') + 2, token.lastIndexOf('"')));
                signal = signals.get(sid);
            } else if (token.contains("origin=\"")) {
                origin = Origintype.valueOf(token.substring(token.indexOf('"') + 1, token.lastIndexOf('"')));
            } else if (token.contains("relType=\"")) {
                if (linkTag == ITimeMLLink.ALink.class) {
                    relType = ALink.valueOf((token.substring(token.indexOf('"') + 1, token.lastIndexOf('"'))));
                } else if (linkTag == ITimeMLLink.SLink.class) {
                    relType = SLink.valueOf((token.substring(token.indexOf('"') + 1, token.lastIndexOf('"'))));
                } else if (linkTag == ITimeMLLink.TLink.class) {
                    relType = TLink.valueOf((token.substring(token.indexOf('"') + 1, token.lastIndexOf('"'))));
                }
            }
        }

        if (linkTag == ITimeMLLink.TLink.class) {
            return new TimeMLLink(linkTag, lid, syntax, firstNode, signal, secondNode,
                    relType, origin);
        } else {
            return new TimeMLLink(linkTag, lid, syntax, firstNode, signal, secondNode,
                    relType);
        }
    }

    /**
     * A method for parsing an instance from a string.
     *
     * @param instanceString a string containing an instance.
     * @return an IInstance object containing it's event instance ID, event,
     * tense, aspect, part of speech, polarity, modality, signal, and cardinality.
     * @throws NullPointerException if null
     * @since jTLEX 1.0
     */
    public ITimeMLInstance parseInstance(String instanceString) {
        int eiid = 0;
        ITimeMLEvent event = null;
        Tense tense = null;
        Aspect aspect = null;
        Pos pos = null;
        Polarity polarity = null;
        String modality = null;
        ITimeMLSignal signal = null;
        String cardinality = null;

        StringTokenizer tok = new StringTokenizer(instanceString);

        while (tok.hasMoreTokens()) {
            String token = tok.nextToken();
            if (token.contains("eiid=\"")) {
                eiid = Integer.parseInt(token.substring(token.indexOf('"') + 3, token.length() - 1));
            } else if (token.contains("eventID=\"")) {
                int eid = Integer.parseInt(token.substring(token.indexOf('"') + 2, token.length() - 1));
                event = events.get(eid);
                
            } else if (token.contains("tense=\"")) {
                tense = Tense.valueOf(token.substring(token.indexOf('"') + 1, token.length() - 1));
            } else if (token.contains("aspect=\"")) {
                aspect = Aspect.valueOf(token.substring(token.indexOf('"') + 1, token.length() - 1));
            } else if (token.contains("pos=\"")) {
                pos = Pos.valueOf((token.substring(token.indexOf('"') + 1, token.lastIndexOf('"'))));
            } else if (token.contains("polarity=\"")) {
                polarity = Polarity.valueOf(token.substring(token.indexOf('"') + 1, token.length() - 1));
            } else if (token.contains("modality=\"")) {
                if (token.indexOf('"') != token.lastIndexOf('"')) {
                    modality = token.substring(token.indexOf('"') + 1, token.lastIndexOf('"')) + " ";
                } else {
                    if (modality == null) {
                        modality = "";
                    }
                    modality += token.substring(token.indexOf('"') + 1, token.length() - 1);
                }
            } else if (token.contains("signalID=\"")) {
                int sid = Integer.parseInt(token.substring(token.indexOf('"') + 4, token.length() - 1));
                signal = signals.get(sid);
            } else if (token.contains("cardinality=\"")) {
                cardinality = token.substring(token.indexOf('"') + 1, token.length() - 1);
            }
        }

        return new TimeMLInstance(eiid, event, tense, aspect, pos, polarity, modality, signal, cardinality);

    }

    /**
     * A method for parsing a signal from a string.
     *
     * @param signalString a string containing a signal.
     * @return an ISignal object containing it's signal ID and string.
     * @throws NullPointerException if null
     * @since jTLEX 1.0
     */
    public ITimeMLSignal parseSignal(String signalString) {
        int sid = 0;
        String signalStr = "";

        StringTokenizer tok = new StringTokenizer(signalString);

        while (tok.hasMoreTokens()) {
            String token = tok.nextToken();
            if (token.contains("sid=\"")) {
                sid = Integer.parseInt(token.substring(token.indexOf('"') + 2, token.length() - 1));
            } else {
                signalStr += token;
            }
        }
        return new TimeMLSignal(sid, signalStr);
    }
}
